from moexalgo import ahdata

def test_ahdata():
    files = ahdata.files()
    assert len(files) > 0
    file = files[0]
    data = ahdata.download(dataset=file.dataset, period=file.year_month)
    assert len(list(data)) > 0


import typing as t
from datetime import date

from pandas import DataFrame

from moexalgo.next import protocols
from moexalgo.next.protocols import CandlePeriod
from ._common import Market as CommonMarket, Ticker as CommonTicker


class Market(CommonMarket, protocols.Market):
    """
    Обобщенный класс `Market` для 'stock'.
    """

    def __init__(self, *args, **kwargs):
        super().__init__("stock", *args, **kwargs)

    def tickers(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        SHARES_DEFAULT_FIELDS = (
            "shortname",  # Краткое наименование
            "lotsize",  # Размер лота
            "decimals",  # Количество знаков после запятой
            "minstep",  # Минимальный шаг цены
            "issuesize",  # Объем выпуска
            "isin",  # Стандартное наименование
            "regnumber",  # Регистрационный номер
            "listlevel",  # Уровень листинга
        )
        match self.market:
            case "shares":
                fields = fields or SHARES_DEFAULT_FIELDS if not all_fields else ()
        return super()._tickers(*fields, all_fields=all_fields, native=native)

    def marketdata(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        SHARES_DEFAULT_FIELDS = (
            "bid",  # Лучшая цена покупки
            "offer",  # Лучшая цена продажи
            "biddeptht",  # Глубина стакана покупки
            "offerdeptht",  # Глубина стакана продажи
            "open",  # Цена открытия
            "high",  # Максимальная цена
            "low",  # Минимальная цена
            "last",  # Цена последней сделки
            "waprice",  # Средневзвешенная цена
            "lasttoprevprice",  # Изменение цены последней сделки к предыдущей
            "numtrades",  # Количество сделок
            "voltoday",  # Объем сделок за день
            "valtoday",  # Объем сделок за день в валюте
            "valtoday_usd",  # Объем сделок за день в долларах
            "openperiodprice",  # Цена открытия периода
            "closingauctionprice",  # Цена закрытия аукциона
            "closingauctionvolume",  # Объем закрытия аукциона
            "issuecapitalization",  # Капитализация
            "updatetime",  # Время обновления
            "systime",  # Время системы
        )
        match self.market:
            case "shares":
                fields = fields or SHARES_DEFAULT_FIELDS if not all_fields else ()
        return super()._marketdata(*fields, all_fields=all_fields, native=native)


class Ticker(CommonTicker, protocols.Ticker):
    """
    Обобщенный класс `Ticker` для 'stock'.
    """

    def info(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> dict[str, t.Any] | DataFrame:
        return super()._info(*fields, all_fields=all_fields, native=native)

    def candles(
        self,
        start: str | date,
        end: str | date,
        period: int | str | CandlePeriod = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._candles(start, end, period, offset=offset, latest=latest, native=native)

    def trades(
        self,
        tradeno: int = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._trades(tradeno, offset=offset, latest=latest, native=native)

    def orderbook(self, native: bool = False) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._orderbook(native)



import re
import statistics
import typing as t
from datetime import datetime, timedelta, date, time

import moexalgo

try:
    import pandas as pd

    type Data = list[dict[str, t.Any]] | pd.DataFrame
except ImportError:
    type Data = list[dict[str, t.Any]]

type Metrics = t.Literal["tradestats", "orderstats", "obstats"]


def _normalize(data: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
    def normalize(item: dict[str, t.Any]):
        if ticker := item.pop('ticker', item.pop('secid', None)):
            end = item.pop('end', None)
            begin = item.pop('begin', None)
            tradedate = item.pop('tradedate', None)
            tradetime = item.pop('tradetime', None)
            if tradedate and tradetime:
                tradedate = date.fromisoformat(tradedate) if isinstance(tradedate, str) else tradedate
                tradetime = time.fromisoformat(tradetime) if isinstance(tradetime, str) else tradetime
                end = datetime.combine(tradedate, tradetime)
                begin = end - timedelta(minutes=5)
            if end and begin:
                begin = datetime.fromisoformat(begin) if isinstance(begin, str) else begin
                end = datetime.fromisoformat(end) if isinstance(end, str) else end
                return dict(ticker=ticker, begin=begin, end=end, **item)
        raise ValueError(f"Wrong item: {item}")

    return list(map(normalize, data))


def _resolve_and_normalize(
        data: list[dict[str, t.Any]]
) -> tuple[Metrics | None, list[dict[str, t.Any]]]:
    metrics = None
    if data:
        match data[0]:
            case {"trades": trades, "vol": vol}:
                if trades and vol:
                    metrics = "tradestats"
            case {"put_orders_b": put_orders, "cancel_orders_b": cancel_orders}:
                if put_orders or cancel_orders:
                    metrics = "orderstats"
            case {"levels_b": levels_b, "levels_s": levels_s}:
                if levels_b or levels_s:
                    metrics = "obstats"
        if metrics:
            data = _normalize(data)
    return metrics, data


def normalize_period(period: int | str):
    if isinstance(period, str):
        if found := re.match("([0-9]+)([a-zA-Z]+)", period):
            number, symbol = found.groups()
            if period == found.group():
                if symbol == "min":
                    number = int(number)
                    if number // 5 == number / 5:
                        return int(number), f"{number}min"
                elif symbol == "h":
                    return int(number) * 60, f"{number}h"
                elif symbol == "d":
                    return int(number) * 60 * 24, f"{number}d"
                elif symbol == "w":
                    return int(number) * 60 * 24 * 7, f"{number}w"
                elif symbol == "m":
                    raise NotImplementedError(f"Period {period} is not supported in this case")
    elif isinstance(period, int):
        if period // 5 == period / 5:
            return period, f"{period}min"
    raise ValueError(f"Wrong period value: {period}")

def normalize_int(value):
    try:
        value = int(float(value))
    except (ValueError, TypeError):
        return None
    return value

def normalize_float(value, decimals=None):
    try:
        value = float(value)
        if decimals is not None:
            value = round(value, decimals)
    except (ValueError, TypeError):
        return None
    return value


def save_(func, items, name):
    items = [item[name] for item in items if item[name] is not None] if items else []
    return func(items) if items else None


def _tradestats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_MEAN = ['pr_std']
    TO_SUM = ['vol', 'val', 'trades', 'trades_b', 'trades_s', 'vol_b', 'vol_s', 'val_b', 'val_s']
    items = [dict((k, normalize_int(v) if k in (
        'trades', 'trades_b', 'trades_s', 'vol', 'vol_b', 'vol_s',
        'sec_pr_open', 'sec_pr_high', 'sec_pr_low', 'sec_pr_close')
    else (normalize_float(v) if k in (
        'pr_std', 'val', 'val_b', 'val_s',
        'pr_open', 'pr_high', 'pr_low', 'pr_close', 'oi_open', 'oi_high', 'oi_low', 'oi_close'
    ) else v)) for k, v in item.items()) for item in items]

    result = None
    for N in range(len(items)):
        tradestat = items[N]
        offset = tradestat.pop('offset')
        if result is None:
            result = tradestat
            result['sec_pr_open'] = offset + tradestat['sec_pr_open']
            result['sec_pr_high'] = offset + tradestat['sec_pr_high']
            result['sec_pr_low'] = offset + tradestat['sec_pr_low']
            result['sec_pr_close'] = offset + tradestat['sec_pr_close']

        if tradestat['pr_high'] > result['pr_high']:
            result['pr_high'] = tradestat['pr_high']
            result['sec_pr_high'] = offset + tradestat['sec_pr_high']
        if tradestat['pr_low'] < result['pr_low']:
            result['pr_low'] = offset + tradestat['pr_low']
            result['sec_pr_low'] = offset + tradestat['sec_pr_low']
        result['pr_close'] = tradestat['pr_close']
        result['sec_pr_close'] = offset + tradestat['sec_pr_close']
        if 'im' in tradestat:
            result['im'] = tradestat['im']

        if 'oi_open' in tradestat:
            if tradestat['oi_high'] > result['oi_high']:
                result['oi_high'] = tradestat['oi_high']
            if tradestat['oi_low'] < result['oi_low']:
                result['oi_low'] = tradestat['oi_low']
            result['oi_close'] = tradestat['oi_close']

    for name in TO_SUM:
        if name in result:
            result[name] = save_(sum, items, name)
    for name in TO_MEAN:
        if name in result:
            try:
                result[name] = save_(statistics.mean, items, name)
            except statistics.StatisticsError:
                pass
    result['pr_std'] = round(result['pr_std'], 4) if result['pr_std'] else None
    result['pr_change'] = round(100 * (result['pr_close'] - result['pr_open']) / result['pr_open'], 4)
    result['disb'] = (round((result['vol_b'] - result['vol_s']) / result['vol'], 2)
                      if result['vol'] > 0 else None)
    result['pr_vwap'] = round(result['val'] / (result['vol'] * lotsize), decimals) if result['vol'] else None
    result['pr_vwap_b'] = round(result['val_b'] / (result['vol_b'] * lotsize), decimals) if result['vol_b'] else None
    result['pr_vwap_s'] = round(result['val_s'] / (result['vol_s'] * lotsize), decimals) if result['vol_s'] else None

    for key, value in tuple(result.items()):
        if 'val' in key:
            result[key] = normalize_float(value, 0)
    return result


def _orderstats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_SUM = ['put_orders_b', 'put_orders_s', 'put_val_b', 'put_val_s', 'put_vol_b', 'put_vol_s',
              'put_vol', 'put_val', 'put_orders', 'cancel_orders_b', 'cancel_orders_s', 'cancel_val_b',
              'cancel_val_s', 'cancel_vol_b', 'cancel_vol_s', 'cancel_vol', 'cancel_val', 'cancel_orders']

    items = [dict((k, normalize_int(v) if k in (
        'put_orders', 'put_vol', 'put_orders_b', 'put_orders_s', 'put_vol_b', 'put_vol_s',
        'cancel_orders', 'cancel_vol', 'cancel_orders_b', 'cancel_orders_s', 'cancel_vol_b', 'cancel_vol_s'
    ) else (normalize_float(v) if k in (
        'put_val', 'put_val_b', 'put_val_s', 'cancel_val', 'cancel_val_b', 'cancel_val_s'
    ) else v)) for k, v in item.items()) for item in items]

    result = None
    for N in range(len(items)):
        orderstat = items[N]
        if result is None:
            result = orderstat
            continue
    for name in TO_SUM:
        if name in result:
            result[name] = save_(sum, items, name)
    result['put_vwap_b'] = (round(result['put_val_b'] / (result['put_vol_b'] * lotsize), decimals)
                            if result['put_vol_b'] else None)
    result['cancel_vwap_b'] = (round(result['cancel_val_b'] / (result['cancel_vol_b'] * lotsize), decimals)
                               if result['cancel_vol_b'] else None)
    result['put_vwap_s'] = (round(result['put_val_s'] / (result['put_vol_s'] * lotsize), decimals)
                            if result['put_vol_s'] else None)
    result['cancel_vwap_s'] = (round(result['cancel_val_s'] / (result['cancel_vol_s'] * lotsize), decimals)
                               if result['cancel_vol_s'] else None)
    for key, value in tuple(result.items()):
        if 'val' in key:
            result[key] = normalize_float(value, 0)
    return result


def _obstats_calculator(items: list[dict[str, t.Any]], lotsize: int, decimals: int) -> dict[str, t.Any]:
    TO_MEAN = [
        'spread_bbo', 'spread_lv10', 'spread_1mio', 'spread_l1', 'spread_l10', 'spread_l2', 'spread_l3', 'spread_l5',
        'spread_l20', 'levels_b', 'levels_s', 'imbalance_val', 'imbalance_vol_bbo', 'imbalance_val_bbo', 'vwap_b_1mio',
        'vwap_s_1mio', 'vol_b', 'vol_s', 'val_b', 'val_s', 'vol_b_l1', 'vol_b_l10', 'vol_b_l2', 'vol_b_l3', 'vol_b_l5',
        'vol_b_l20', 'vol_s_l1', 'vol_s_l10', 'vol_s_l2', 'vol_s_l20', 'vol_s_l3', 'vol_s_l5', 'micro_price',
        'mid_price'
    ]

    items = [dict((k, normalize_int(v) if k in (
        'levels_b', 'levels_s', 'vol_b', 'vol_s', 'vol_b_l1', 'vol_b_l10', 'vol_b_l2', 'vol_b_l3',
        'vol_b_l5', 'vol_b_l20', 'vol_s_l1', 'vol_s_l10', 'vol_s_l2', 'vol_s_l20', 'vol_s_l3', 'vol_s_l5'
    ) else (normalize_float(v) if k in (
        'spread_bbo', 'spread_lv10', 'spread_1mio', 'spread_l1', 'spread_l10', 'spread_l2', 'spread_l3',
        'spread_l5', 'spread_l20', 'imbalance_vol_bbo', 'imbalance_val_bbo', 'vwap_b_1mio', 'vwap_s_1mio',
        'val_b', 'val_s', 'micro_price', 'mid_price', 'imbalance_val'
    ) else v)) for k, v in item.items()) for item in items]

    result = None
    for N in range(len(items)):
        obstat = items[N]
        if result is None:
            result = obstat
            continue
    for name in TO_MEAN:
        if name in result:
            try:
                result[name] = save_(statistics.mean, items, name)
            except statistics.StatisticsError:
                pass
    for vwap, val, vol in [
        ('vwap_b', 'val_b', 'vol_b'),
        ('vwap_b_l1', 'val_b_l1', 'vol_b_l1'),
        ('vwap_b_l2', 'val_b_l2', 'vol_b_l2'),
        ('vwap_b_l3', 'val_b_l3', 'vol_b_l3'),
        ('vwap_b_l5', 'val_b_l5', 'vol_b_l5'),
        ('vwap_b_l10', 'val_b_l10', 'vol_b_l10'),
        ('vwap_b_l20', 'val_b_l20', 'vol_b_l20'),
        ('vwap_s', 'val_s', 'vol_s'),
        ('vwap_s_l1', 'val_s_l1', 'vol_s_l1'),
        ('vwap_s_l2', 'val_s_l2', 'vol_s_l2'),
        ('vwap_s_l3', 'val_s_l3', 'vol_s_l3'),
        ('vwap_s_l5', 'val_s_l5', 'vol_s_l5'),
        ('vwap_s_l10', 'val_s_l10', 'vol_s_l10'),
        ('vwap_s_l20', 'val_s_l20', 'vol_s_l20'),

    ]:
        if vol in result and val in result:
            result[vwap] = round(result[val] / (result[vol] * lotsize), decimals) if result[vol] else None

    for key, value in tuple(result.items()):
        if 'spread' in key:
            result[key] = normalize_float(value, 1)
        elif 'imbalance' in key:
            result[key] = normalize_float(value, 2)
        elif 'val' in key:
            result[key] = normalize_float(value, 0)
        elif 'vol' in key or 'levels' in key:
            result[key] = normalize_int(value)
        elif 'vwap' in key or 'price' in key:
            result[key] = normalize_float(value, decimals)
    return result


class Resampler:

    def __init__(self, metrics: Metrics, ticker: str, from_date: date, period: int, *args):
        match metrics:
            case "tradestats":
                self._calculator = _tradestats_calculator
            case "orderstats":
                self._calculator = _orderstats_calculator
            case "obstats":
                self._calculator = _obstats_calculator
            case _:
                raise NotImplementedError(f"Resampler is not support '{metrics}' metrics")

        self._ticker = ticker
        self._accum = list()
        self._calculator_args = args
        self._intervals = self._intervals_gen(from_date, period)
        self._begin, self._end = next(self._intervals)

    def __call__(self, item: dict[str, t.Any] = None) -> dict[str, t.Any] | None:
        result = None
        if item is not None:
            begin = item.pop('begin')
            end = item.pop('end')
            while not (self._begin <= begin and end <= self._end):
                if result is None:
                    result = self._get_result()
                self._begin, self._end = next(self._intervals)
            item['offset'] = (begin - self._begin).total_seconds()
            self._accum.append(item)
        else:
            result = self._get_result()
        return result

    def _get_result(self):
        result = None
        if self._accum:
            result = dict(ticker=self._ticker, begin=self._begin, end=self._end,
                          **self._calculator(self._accum, *self._calculator_args))
            result.pop('offset', None)
            self._accum.clear()
        return result

    @staticmethod
    def _intervals_gen(from_date: date, period: int):
        begin = datetime.combine(from_date, time(0))
        end = begin + timedelta(minutes=period)
        while begin.year <= date.today().year:
            yield begin, end
            begin = end
            end = begin + timedelta(minutes=period)

    @classmethod
    def it(cls, metrics, data: list[dict[str, t.Any]], period: int):
        resamplers = dict()
        ticker_args = dict(
            (item['SECID'], (item.get('LOTSIZE', item.get('LOTVOLUME', -1)), item['DECIMALS']))
            for item in sum(list([a for a in item]
                                 for item in [market.tickers(use_dataframe=False)
                                              for market in [moexalgo.Market(market)
                                                             for market in ['EQ', 'FX', 'FO']]]), start=[])
        )
        for item in data:
            ticker = item.pop('ticker')
            if ticker in ticker_args:
                resampler = resamplers.setdefault(ticker,
                                                  cls(metrics, ticker, item['begin'].date(), period, *ticker_args[ticker]))
                if result := resampler(item):
                    yield result
        for resampler in resamplers.values():
            if result := resampler():
                yield result


def resample(data: Data, period: int | str) -> Data:
    if not isinstance(data, list):
        data = data.to_dict('records')
        used_fd = True
    else:
        used_fd = False
    period, _ = normalize_period(period)
    metrics, data = _resolve_and_normalize(data)
    if metrics is None:
        raise NotImplementedError("Resample is not implement for this data")
    result = list(Resampler.it(metrics, data, period))
    return pd.DataFrame(result) if used_fd else result



import typing as t
from datetime import date

from pandas import DataFrame

from moexalgo.next import protocols
from moexalgo.next.protocols import CandlePeriod
from ._common import Market as CommonMarket, Ticker as CommonTicker


class Market(CommonMarket, protocols.Market):
    """
    Обобщенный класс `Market` для 'futures'.
    """

    def __init__(self, *args, **kwargs):
        super().__init__("futures", *args, **kwargs)

    def tickers(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._tickers(*fields, all_fields=all_fields, native=native)

    def marketdata(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._marketdata(*fields, all_fields=all_fields, native=native)


class Ticker(CommonTicker, protocols.Ticker):
    """
    Обобщенный класс `Ticker` для 'futures'.
    """

    def info(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> dict[str, t.Any] | DataFrame:
        return super()._info(*fields, all_fields=all_fields, native=native)

    def candles(
        self,
        start: str | date,
        end: str | date,
        period: int | str | CandlePeriod = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._candles(start, end, period, offset=offset, latest=latest, native=native)

    def trades(
        self,
        tradeno: int = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._trades(tradeno, offset=offset, latest=latest, native=native)

    def orderbook(self, native: bool = False) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        return super()._orderbook(native)



import pytest

from moexalgo.next import Market, Ticker


def test_next_market_bonds(apikey):
    bonds = Market("bonds")
    assert list(bonds.tickers())
    with pytest.raises(AttributeError):
        assert bonds.info()


def test_next_ticker_bonds(apikey):
    su25085rmfs0 = Ticker("SU25085RMFS0")
    assert su25085rmfs0.info()
    with pytest.raises(AttributeError):
        assert su25085rmfs0.tickers()

    bond = Ticker("SU26212RMFS9")
    candles = bond.candles("2025-09-19", "2025-09-19", "10min")
    assert list(candles)

    trades = bond.trades()
    assert list(trades)



import sys
from datetime import date
from types import ModuleType

import moexalgo.next.engines.currency
import moexalgo.next.engines.futures
import moexalgo.next.engines.stock
from moexalgo.next import engines
from moexalgo.next import protocols
from moexalgo.session import Session


def Market(name: str, board: str = None) -> protocols.Market:
    """
    Представление раздела биржевого рынка.

    Parameters
    ----------
    name : str
        Название рынка или его символическое имя.
    board : str, optional
        Идентификатор, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указано, автоматически определятся первичный идентификатор на основе общих правил.
    """
    RESOLVE_MAP = {
        (engines.currency, "selt", "CETS"): ["selt", "currency", "fx"],
        (engines.stock, "shares", "TQBR"): ["shares", "stocks", "equity", "eq"],
        (engines.stock, "index", "SNDX"): ["index"],
        (engines.stock, "bonds", "TQOB"): ["bonds"],
        (engines.futures, "forts", "RFUD"): ["futures", "forts", "fo"],
        (engines.futures, "options", "ROPD"): ["options"],
    }

    def resolve(name: str, boardid: str | None) -> tuple[ModuleType, str, str]:
        for (module, market, default_boardid), aliases in RESOLVE_MAP.items():
            if name in aliases:
                return module, market, boardid or default_boardid
        raise LookupError(f"Unrecognized market' name '{name}'")

    module, name, boardid = resolve(name.lower(), board.upper() if board else None)
    return getattr(module, "Market")(name, boardid)


def Ticker(name: str, board: str = None) -> protocols.Ticker:
    """
    Представление конкретного финансового инструмента.

    Parameters
    ----------
    name : str
        Тикер финансового инструмента, например "GAZP" для акций Газпрома.
    board : str, optional
        Идентификатор, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указано, автоматически определятся первичный идентификатор на основе общих правил.
    """

    def resolve(secid: str, boardid: str | None) -> tuple[str, str, str, int, bool]:
        with Session() as client:
            data = client.get_objects(
                f"securities/{secid}",
                lambda data: [dict(zip(data["boards"]["columns"], row)) for row in data["boards"]["data"]],
            )
        if found := [
            item
            for item in data
            if item["secid"] == secid and (item["boardid"] == boardid if boardid is not None else item["is_primary"])
        ]:
            match found[0]:
                case {"boardid": boardid, "market": market, "decimals": decimals, "listed_till": listed_till}:
                    return market, boardid, secid, decimals, date.fromisoformat(listed_till) < date.today()
        raise LookupError(f"Unrecognized ticker' name '{secid}'")

    market, boardid, secid, decimals, delisted = resolve(name.upper(), board.upper() if board is not None else None)
    market = Market(market, boardid)
    return getattr(sys.modules[market.__module__], "Ticker")(market, boardid, secid, decimals, delisted)



import sys
import typing as t
from datetime import date
from enum import Enum
from types import ModuleType

if "ipykernel" in sys.modules:
    from pandas import DataFrame
else:
    type DataFrame = t.Any


class CandlePeriod(Enum):
    """
    Возможные временные интервалы для свечей.
    """

    ONE_MINUTE = 1
    TEN_MINUTES = 10
    ONE_HOUR = 60
    ONE_DAY = 24
    ONE_WEEK = 7
    ONE_MONTH = 31


class Market(t.Protocol):
    """
    Протокол объекта `moexalgo.Market`
    """

    module: ModuleType

    def tickers(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Информация о всех инструментах рынка.

        Parameters
        ----------
        fields : tuple[str]
            Поля для отображения
        all_fields : bool, optional
            Если флаг выставлен в `True` выводятся все имеющиеся поля.
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается итератор словарей.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """

    def marketdata(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Статистическая информация о всех инструментах рынка.

        Parameters
        ----------
        fields : tuple[str]
            Поля для отображения
        all_fields : bool, optional
            Если флаг выставлен в `True` выводятся все имеющиеся поля.
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается итератор словарей.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """


class Ticker(t.Protocol):
    """
    Протокол объекта `moexalgo.Ticker`
    """

    def info(self, *fields: str, all_fields: bool = False, native: bool = False) -> dict[str, t.Any] | DataFrame:
        """
        Информация об инструменте.

        Parameters
        ----------
        fields : tuple[str]
            Поля для отображения
        all_fields : bool, optional
            Если флаг выставлен в `True` выводятся все имеющиеся поля.
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается словарь.

        Returns
        -------
        return : dict | pd.DataFrame
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """

    def candles(
        self,
        start: str | date,
        end: str | date,
        period: int | str | CandlePeriod = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает свечи инструмента по заданным параметрам.

        Parameters
        ----------
        start : Union[str, date]
            Дата начала диапазона выдачи данных. (`start` может быть равен `end`, тогда вернутся записи за один день)
        end : Union[str, date]
            Дата конца диапазона выдачи данных.
        period : Union[str, int, CandlePeriod], optional
            Период свечи, by default None.

            Если `int`, то интерпретируется в следующем формате:
            - 1 - 1 минута
            - 10 - 10 минут
            - 60 - 1 час
            - 24 - 1 день
            - 7 - 1 неделя
            - 31 - 1 месяц

            Если `str`, то интерпретируется в следующем формате:
            - '1min' - 1 минута
            - '10min' - 10 минут
            - '1h' - 1 час
            - '1d' - 1 день
            - '1w' - 1 неделя
            - '1m' - 1 месяц

        offset : int, optional
            Начальная позиция в последовательности записей, by default 0.
        latest : bool, optional
            Включает режим выдачи последних `latest` записей в наборе, by default False.
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается итератор словарей.

        Returns
        -------
        return : t.Iterable[dict[str, t.Any]] | DataFrame
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """

    def trades(
        self,
        tradeno: int = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает сделки за последний день или начиная с заданного `tradeno`.

        Parameters
        ----------
        tradeno : int, optional
            Номер сделки с которого выдаются данные, если не задано, то с начала дня.
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается итератор словарей.

        Returns
        -------
        return : t.Iterable[dict[str, t.Any]] | DataFrame
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """

    def orderbook(
        self,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        """
        Возвращает текущий стакан лучших цен.

        Parameters
        ----------
        native : bool, optional
            Если флаг выставлен в `True` всегда возвращается итератор словарей.

        Returns
        -------
        return : t.Iterable[dict[str, t.Any]] | DataFrame
            При использовании в среде Jupiter Notebook по умалчиванию данные возвращаются в виде `DataFrame`.
            Используйте флаг `native` что бы отменить такое поведение.
        """



from __future__ import annotations

from datetime import date
from typing import Optional, Union

from moexalgo import session
from moexalgo.metrics import prepare_market_request, dataclass_it, pandas_frame
from moexalgo.session import Session
from moexalgo.utils import result_deserializer, pd

_AVAILABLE = {
    "index": dict(),
    "shares": dict(),
    "selt": dict(),
    "forts": dict(),
    "bonds": dict(),
}
_ALIASES = {
    "index": ("engines/stock/markets/index", "SNDX"),
    "shares": ("engines/stock/markets/shares", "TQBR"),
    "stock": ("engines/stock/markets/shares", "TQBR"),
    "stocks": ("engines/stock/markets/shares", "TQBR"),
    "EQ": ("engines/stock/markets/shares", "TQBR"),
    "selt": ("engines/currency/markets/selt", "CETS"),
    "FX": ("engines/currency/markets/selt", "CETS"),
    "currency": ("engines/currency/markets/selt", "CETS"),
    "forts": ("engines/futures/markets/forts", "RFUD"),
    "FO": ("engines/futures/markets/forts", "RFUD"),
    "futures": ("engines/futures/markets/forts", "RFUD"),
}


def market_for(secid: str, boardid: str, cs: Session = None) -> Optional[Market]:
    """
    Возвращает рынок, на котором торгуется заданный инструмент.

    Parameters
    ----------
    secid : str
        Уникальный идентификатор инструмента.
    boardid : str
        Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
    cs : Session, optional
        Клиентская сессия, если используется.

    Returns
    -------
    return : Optional[Market]
        Экземпляр класса Market, представляющий рынок, на котором торгуется заданный инструмент.
    """
    for _, boards in _AVAILABLE.items():
        if market := boards.get(boardid):
            if market.securities_for(secid, cs):
                return market


class Market:
    """
    Представление конкретного раздела биржевого рынка.

    Больше информации о FUTOI можно найти на странице: https://moexalgo.github.io/des/futoi/

    Этот класс предоставляет доступ к информации о биржевом рынке, а также позволяет
    оперировать данными конкретного рынка по его уникальному идентификатору.

    Parameters
    ----------
    name : str
        Название рынка или его символическое имя.
    boardid : str, optional
        Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
        Если не указан, класс попытается автоматически определить идентификатор на основе общих правил.

    Methods
    -------
    tickers : Callable[[Session, bool], Union[List[Dict[str, Any]], pandas.DataFrame]]
        Возвращает список всех инструментов рынка.

    marketdata : Callable[[Session, bool], Union[List[Dict[str, Any]], pandas.DataFrame]]
        Возвращает статистическую информацию о всех инструментах рынка.

    tradestats : Callable[[Session, bool], Union[Iterator[TradeStat], pandas.DataFrame]]
        Возвращает метрики `TradeStat` по заданным параметрам.

    orderstats : Callable[[Session, bool], Union[Iterator[OrderStat], pandas.DataFrame]]
        Возвращает метрики `OrderStat` по заданным параметрам.

    obstats : Callable[[Session, bool], Union[Iterator[ObStat], pandas.DataFrame]]
        Возвращает метрики `ObStat` по заданным параметрам.

    futoi : Callable[[Session, bool], Union[Iterator[FUTOI], pandas.DataFrame]]
        Возвращает метрики `FUTOI` по заданным параметрам.

    Returns
    -------
    return : Market
        Экземпляр класса Market, представляющий указанный рынок.

    Raises
    ------
    NotImplementedError
        Вызывается, если рынок с указанным названием не поддерживается.

    Example
    -------
    .. code-block:: python

        # Получение информации об акции
        >>> try:
        >>>     market_instance = Market("index", "MOEX")
        >>>     print(market_instance)
        >>>     market_instance._ensure_loaded()
        >>> except NotImplementedError:
        >>>     print("Рынок не поддерживается.")
    """

    _name: str
    _path: str
    _pref: str
    _boardid: str
    _fields: dict[str, dict[str, dict]] = None
    _values: dict[str, dict[str, dict]] = None
    _delisted: list[str] = None
    _LIMIT = 25_000

    def __new__(cls, name: str, boardid: str = None) -> Market:
        """
        Создает новый экземпляр класса Market.

        Parameters
        ----------
        name : str
            Название рынка или его символическое имя.
        boardid : str, optional
            Идентификатор рынка, указывающий на специфическую торговую площадку или сегмент рынка.
            Если не указан, класс попытается автоматически определить идентификатор на основе общих правил.

        Returns
        -------
        return : Market
            Экземпляр класса Market, представляющий указанный рынок.

        Raises
        ------
        NotImplementedError
            Вызывается, если рынок с указанным названием не поддерживается.

        Example
        -------
        .. code-block:: python

            # Получение информации об акции
            >>> market_instance = Market("index", "MOEX")
            >>> print(market_instance)
        """
        if "/" not in name:
            path, default_boardid = _ALIASES.get(name, (None, None))
            if path is None:
                raise NotImplementedError(f"Market '{name}' is not implemented.")
            boardid = boardid or default_boardid
            name = path.split("/")[-1]
        else:
            path = name
            name = name.split("/")[-1]

        if name not in _AVAILABLE:
            raise NotImplementedError(f"Market {name} is not supported")

        market = _AVAILABLE.setdefault(name, dict())
        prefs = [k for k, v in _ALIASES.items() if v[0].startswith(path) and len(k) == 2]
        pref = prefs[0].lower() if prefs else None
        if boardid not in market:
            market[boardid] = super().__new__(cls)
            market[boardid]._name = name
            market[boardid]._path = path
            market[boardid]._pref = pref
            market[boardid]._boardid = boardid

        return market[boardid]

    def __repr__(self):
        return f"{self.__class__.__name__}('{self._name}/{self._boardid}')"

    def _ensure_loaded(self, cs: Session = None) -> None:
        """
        Проверяет, загружены ли данные о рынке и его инструментах.
        Если данные не загружены, загружает их.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.

        Returns
        -------
        return : None
        """
        if self._fields is None or self._values is None:
            with Session(cs or session.default) as client:
                self._fields = client.get_objects(
                    f"{self._path}/boards/{self._boardid}/securities/columns",
                    lambda data: result_deserializer(data, key=lambda item: item["name"]),
                )

                self._values = client.get_objects(
                    f"{self._path}/boards/{self._boardid}/securities/",
                    lambda data: result_deserializer(data, key=lambda item: item["SECID"]),
                )

    def _is_delisted(self, secid: str, cs: Session = None):
        raise NotImplementedError("Deprecated")
        # self._ensure_loaded(cs)
        # if self._delisted is None:
        #     market = self._name
        #     engine = self._path.split('/')[1]
        #     "&group_by=type&group_by_filter=preferred_share&iss.only=securities&securities.columns=secid"
        #     with Session(cs or session.default) as client:
        #         data = data_gen(cs, f'securities/',
        #                         {'engine': engine, 'market': market, 'is_trading': 0,
        #                          'iss.only': 'securities', 'securities.columns': 'secid'},
        #                         0, 10000, 'securities')
        #         self._delisted = [row['secid'] for row in data]
        # return secid in self._delisted

    def _ticker_info(self, secid: str, cs: Session = None) -> dict[str, dict[str, dict]]:
        """
        Возвращает информацию о заданном инструменте.

        Parameters
        ----------
        secid : str
            Уникальный идентификатор инструмента.
        cs : Session, optional
            Клиентская сессия, если используется.

        Returns
        -------
        return : Dict[str, Any]
            Информация о заданном инструменте.
        """
        raise NotImplementedError("Deprecated")
        # self._ensure_loaded(cs)
        # marketdata = self._values['marketdata'].get(secid)
        # securities = self._values['securities'].get(secid)
        # if securities or marketdata:
        #     return dict(securities=securities, marketdata=marketdata)
        # if self._is_delisted(secid):
        #     return dict(securities=None, marketdata=None)

    def _normalize_row(self, row: dict[str, dict], fields: tuple[str]) -> dict[str, dict]:
        """
        Нормализует строку данных о статистике инструмента.

        Parameters
        ----------
        row : Dict[str, Any]
            Строка данных о статистике инструмента.
        fields : Tuple[str]
            Поля, которые необходимо оставить в строке данных.

        Returns
        -------
        return : Dict[str, Any]
            Нормализованная строка данных о статистике инструмента.
        """
        return dict(ticker=row["SECID"], **{key.lower(): value for key, value in row.items() if key in fields})

    def _get_data(
        self, option: str, cs: Session = None, use_dataframe: bool = True, fields: tuple[str] = None
    ) -> Union[list[dict], pd.DataFrame]:
        """
        Возвращает данные о рынке по заданному параметру.

        Parameters
        ----------
        option : str
            Параметр, по которому необходимо получить данные.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.
        fields : Tuple[str], optional
            Поля, которые необходимо оставить в строке данных.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        self._ensure_loaded(cs)
        if use_dataframe:
            return pd.DataFrame([self._normalize_row(row, fields) for row in self._values[option].values()])
        else:
            return list(self._values[option].values())

    def tickers(self, cs: Session = None, use_dataframe: bool = True) -> Union[list[dict], pd.DataFrame]:
        """
        Возвращает список всех инструментов рынка.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        fields = (
            "SHORTNAME",  # Краткое наименование
            "LOTSIZE",  # Размер лота
            "DECIMALS",  # Количество знаков после запятой
            "MINSTEP",  # Минимальный шаг цены
            "ISSUESIZE",  # Объем выпуска
            "ISIN",  # Стандартное наименование
            "REGNUMBER",  # Регистрационный номер
            "LISTLEVEL",  # Уровень листинга
        )

        return self._get_data("securities", cs, use_dataframe, fields)

    def marketdata(self, cs: Session = None, use_dataframe: bool = True) -> Union[list[dict], pd.DataFrame]:
        """
        Возвращает статистическую информацию о всех инструментах рынка.

        Parameters
        ----------
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе список.

        Returns
        -------
        return : Union[list[dict], pd.DataFrame]
            Объекты типа List или `pd.DataFrame`.
        """
        fields = (
            "BID",  # Лучшая цена покупки
            "OFFER",  # Лучшая цена продажи
            "BIDDEPTHT",  # Глубина стакана покупки
            "OFFERDEPTHT",  # Глубина стакана продажи
            "OPEN",  # Цена открытия
            "HIGH",  # Максимальная цена
            "LOW",  # Минимальная цена
            "LAST",  # Цена последней сделки
            "WAPRICE",  # Средневзвешенная цена
            "LASTTOPREVPRICE",  # Изменение цены последней сделки к предыдущей
            "NUMTRADES",  # Количество сделок
            "VOLTODAY",  # Объем сделок за день
            "VALTODAY",  # Объем сделок за день в валюте
            "VALTODAY_USD",  # Объем сделок за день в долларах
            "OPENPERIODPRICE",  # Цена открытия периода
            "CLOSINGAUCTIONPRICE",  # Цена закрытия аукциона
            "CLOSINGAUCTIONVOLUME",  # Объем закрытия аукциона
            "ISSUECAPITALIZATION",  # Капитализация
            "UPDATETIME",  # Время обновления
            "SYSTIME",  # Время системы
        )

        return self._get_data("marketdata", cs, use_dataframe, fields)

    def _prepare_metric(
        self,
        metric: str,
        metric_type: str = None,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
        limit: int = None,
    ) -> Union[iter, pd.DataFrame]:
        """
        Подготавливает запрос к рынку по заданным параметрам.

        Parameters
        ----------
        metric : str
            Метрика, по которой необходимо получить данные.
        metric_type : str, optional
            Тип метрики, by default None
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Iterator[TradeStat] | pd.DataFrame
            Итератор или `pd.DataFrame` метрик.
        """
        metrics_it = prepare_market_request(
            f"{metric_type}/{metric}", cs, date_=date, latest=latest, offset=offset, limit=limit or 50_000
        )
        return pandas_frame(metrics_it) if use_dataframe else dataclass_it(metrics_it)

    def tradestats(
        self,
        *,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `TradeStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если True, то возвращает pd.DataFrame, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или pd.DataFrame метрик `TradeStat`.
        """

        return self._prepare_metric("tradestats", self._pref, date, latest, offset, cs, use_dataframe)

    def orderstats(
        self,
        *,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `OrderStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или pd.DataFrame метрик `OrderStat`.
        """

        return self._prepare_metric("orderstats", self._pref, date, latest, offset, cs, use_dataframe)

    def obstats(
        self,
        *,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `ObStat` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `ObStat`.
        """
        return self._prepare_metric("obstats", self._pref, date, latest, offset, cs, use_dataframe)

    def futoi(
        self,
        *,
        date: Union[str, date] = None,
        # latest: bool = None,
        # offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `FUTOI` по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame` метрик `FUTOI`.

        Raises
        ------
        NotImplementedError
            Вызывается, если `FUTOI` не поддерживается для данного рынка.
        """
        if not self._path.startswith("engines/futures/markets"):
            raise NotImplementedError("FUTOI is not implemented for this market")

        return self._prepare_metric(
            "FUTOI", self._pref, date, latest=None, offset=None, cs=cs, use_dataframe=use_dataframe, limit=-1
        )

    def alerts(
        self,
        *,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает MegaAlert (оповещение об аномальной рыночной активности) по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame`.
        """
        if not self._path.startswith("engines/stock/markets"):
            raise NotImplementedError
        return self._prepare_metric("alerts", self._pref, date, latest, offset, cs, use_dataframe)

    def hi2(
        self,
        *,
        date: Union[str, date] = None,
        latest: bool = None,
        offset: int = None,
        cs: Session = None,
        use_dataframe: bool = True,
    ) -> Union[iter, pd.DataFrame]:
        """
        Возвращает метрики `Hi2` (индекс рыночной концентрации) по заданным параметрам.

        Parameters
        ----------
        date : Union[str, date], optional
            Дата данных. Если не указано, данные выдаются за сегодняшнее число.
        latest : bool, optional
            Включает режим выдачи последних записей в наборе.
        offset : int, optional
            Начальная позиция в последовательности записей.
        cs : Session, optional
            Клиентская сессия, если используется.
        use_dataframe : bool, optional
            Изменяет тип возвращаемого объекта, by default `True`.
            Если `True`, то возвращает `pd.DataFrame`, иначе итератор.

        Returns
        -------
        return : Union[iter, pd.DataFrame]
            Итератор или `pd.DataFrame`.
        """
        return self._prepare_metric("hi2", self._pref, date, latest, offset, cs, use_dataframe)



import moexalgo
import moexalgo.tools


def test_resample_trades():
    moex = moexalgo.Ticker('MOEX')
    data_5m = moex.obstats(start="2025-01-01", end="2025-01-31")
    data_1d = moexalgo.tools.resample(data_5m, period="1d")
    assert data_1d




import typing as t
from collections.abc import Iterable
from datetime import date

from moexalgo.candles import normalize_period
from moexalgo.metrics import prepare_from_till_dates, calc_offset_limit
from moexalgo.next.protocols import DataFrame, CandlePeriod
from moexalgo.session import Session

pd = None
if not isinstance(DataFrame, t.TypeAliasType):
    import pandas as pd


class Market:
    """
    Обобщенный класс `Market`
    """

    def __init__(self, engine: str, market: str, boardid: str):
        self.market = market
        self.boardid = boardid
        self.engine = engine

    def _tickers(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
        default_fields: list[str] = None,
    ) -> Iterable[dict[str, t.Any]] | DataFrame:
        if not fields:
            if default_fields:
                fields = default_fields
            else:
                all_fields = True

        def fetch():
            with Session() as client:
                data = client.get_objects(
                    f"engines/{self.engine}/markets/{self.market}/boards/{self.boardid}/securities",
                    lambda data: [dict(zip(data["securities"]["columns"], row)) for row in data["securities"]["data"]],
                )
                for item in data:
                    yield normalize_item(item, *fields)

        return result_adapter(fetch(), native)

    def _marketdata(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
        default_fields: list[str] = None,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        if not fields:
            if default_fields:
                fields = default_fields
            else:
                all_fields = True

        def fetch():
            with Session() as client:
                data = client.get_objects(
                    f"engines/{self.engine}/markets/{self.market}/boards/{self.boardid}/securities",
                    lambda data: [dict(zip(data["marketdata"]["columns"], row)) for row in data["marketdata"]["data"]],
                )
                for item in data:
                    yield normalize_item(item, *fields)

        return result_adapter(fetch(), native)


class Ticker:
    """
    Обобщенный класс `Ticker`
    """

    def __init__(self, market: Market, boardid: str, secid: str, decimals: int, delisted: bool):
        self.market = market
        self.boardid = boardid
        self.secid = secid
        self.decimals = decimals
        self.delisted = delisted

    def _info(
        self,
        *fields: str,
        all_fields: bool = False,
        native: bool = False,
        default_fields: list[str] = None,
    ) -> dict[str, t.Any] | DataFrame:
        if not fields:
            if default_fields:
                fields = default_fields
            else:
                all_fields = True

        def fetch():
            with Session() as client:
                data = client.get_objects(
                    f"securities/{self.secid}",
                    lambda data: [dict(zip(data["boards"]["columns"], row)) for row in data["boards"]["data"]],
                )
                for item in data:
                    if item["secid"] == self.secid:
                        yield normalize_item(item, *fields)

        return result_adapter(fetch(), native)

    def _candles(
        self,
        start: str | date,
        end: str | date,
        period: int | str | CandlePeriod = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        interval_seconds = normalize_period(period)
        from_date, till_date = prepare_from_till_dates(start, end)
        options = {"from": from_date, "till": till_date, "interval": interval_seconds}
        offset, limit = calc_offset_limit(offset, None)
        if latest:
            options["iss.reverse"] = True
            limit = 1
        options = dict(options, offset=offset, limit=limit)

        def fetch():
            for item in fetch_section(
                self.market.engine,
                self.market.market,
                self.boardid,
                self.secid,
                "candles",
                **options,
            ):
                yield normalize_item(item)

        return result_adapter(fetch(), native)

    def _trades(
        self,
        tradeno: int = None,
        *,
        offset: int = 0,
        latest: bool = False,
        native: bool = False,
    ) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        options = {"tradeno": tradeno}
        offset, limit = calc_offset_limit(offset, None)
        if latest:
            options["iss.reverse"] = True
            limit = 1
        options = dict(options, offset=offset, limit=limit)

        def fetch():
            for item in fetch_section(
                self.market.engine,
                self.market.market,
                self.boardid,
                self.secid,
                "trades",
                **options,
            ):
                yield normalize_item(item)

        return result_adapter(fetch(), native)

    def _orderbook(self, native: bool = False) -> t.Iterable[dict[str, t.Any]] | DataFrame:
        options = dict(offset=0, limit=-1)

        def fetch():
            for item in fetch_section(
                self.market.engine,
                self.market.market,
                self.boardid,
                self.secid,
                "orderbook",
                **options,
            ):
                yield normalize_item(item)

        return result_adapter(fetch(), native)


def normalize_item(item: dict[str, t.Any], *fields: tuple[str | tuple[str, str]]) -> dict[str, t.Any]:
    names = {"secid": "ticker", "boardid": "board"}
    if fields:
        names.update(dict(((item, item) if isinstance(item, str) else item) for item in fields))

    def normalize_name(name: str) -> str:
        name = name.lower()
        return names.get(name, name)

    pre_result = [(normalize_name(k), v) for k, v in item.items()]
    names = tuple(names.values())
    return dict((k, v) for k, v in pre_result if len(fields) == 0 or k in names)


def fetch_section(
    engine: str,
    market: str,
    boardid: str,
    secid: str,
    section: str,
    *,
    limit: int,
    offset: int = 0,
    **options: t.Any,
) -> t.Iterable[dict[str, t.Any]]:
    start = offset
    with Session() as client:
        while True:
            options["start"] = start
            if data := client.get_objects(
                f"engines/{engine}/markets/{market}/boards/{boardid}/securities/{secid}/{section}",
                lambda data: [dict(zip(data[section]["columns"], row)) for row in data[section]["data"]],
                **options,
            ):
                for item in data:
                    yield item
                    start += 1
                if (start - offset) < limit and limit > 0:
                    continue
            break


def result_adapter(result: dict | t.Iterable[dict], native: bool) -> dict | t.Iterable[dict] | DataFrame | None:
    """Преобразует результат в `pandas.DataFrame` если применимо."""
    use_dataframe = not native and pd is not None
    if use_dataframe:
        if result is None:
            return None
        result = pd.DataFrame([result] if isinstance(result, dict) else list(result))
    return result
